from .Ackley import Ackley
from .Bukin import Bukin
from .Branin import Branin
from .Dropwave import DropWave
from .Eggholder import EggHolder
from .Griewank import Griewank
from .Hartmann6d import Hartmann6d
from .Holdertable import HolderTable
from .Productsines import ProductSines
from .Rastrigin import Rastrigin
from .Rosenbrock import Rosenbrock
from .Shubert import Shubert
from .Sphere import Sphere
from .Trid import Trid
from .Noise import Noise
from .Zero import Zero
from .Sin import Sin
from .AbsMax import AbsMax

Base_func_list = [
    Ackley,
    Bukin,
    Branin,
    DropWave,
    EggHolder,
    Griewank,
    Hartmann6d,
    HolderTable,
    ProductSines,
    Rastrigin,
    Rosenbrock,
    Shubert,
    Sphere,
    Trid,
    Noise,
    Zero,
    Sin,
    AbsMax,
]
